import hmac
import pytest
import os

from hashlib import blake2b, sha512

from bip32 import BIP32
from bip_utils import Bech32Encoder

from nacl import bindings

from mimblewimble.wallet import Wallet
from mimblewimble.mnemonic import Mnemonic


# this wallet contains no funds, was generated using grin-wallet 5.0.3 for test purposes,
# do NOT send any funds to this wallet and if you do, consider them lost as the seed is publicly disclosed

seed = {
    'encrypted_seed': '839773da8062af7dc51714fd98a7f9a72750e17aa54541d5317b5ea1be5c5751db85497aa630380dd984e2ecd603ae0b',
    'salt': '356f045acf2b2787',
    'nonce': 'b5a490e1c942e6a5147bb740'
}

password = 'grinventions'

recovery_phrase = 'sign interest obtain raw window monster jump bring nice crunch toward grunt prosper recycle sphere battle mother fold reject velvet emotion similar romance govern'

accounts = [
    (
        'default',
        'm/0/1/0', # core wallet shows m/0/0
        'grin14kgku7l5x6te3arast3p59zk4rteznq2ug6kmmypf2d6z8md76eqg3su35'
    ),
    (
        'alice',
        'm/1/1/0', # core wallet shows m/1/0
        'grin1uqan8sf49yf0369ezef9jhl25jll9fc8xc5wjkcg0w6nv6v85v2sp4wgwy'
    ),
    (
        'bob',
        'm/2/1/0', # core wallet shows m/2/0
        'grin1guszgjsjlt9vrppu42l03xx080epzzvse5nev3nvdh632explc0sj8ylja'
    ),
]

def test_reflexivity():
    random_password = os.urandom(32).hex()
    w = Wallet.initialize()

    slatepack_address = w.getSlatepackAddress()
    recovery_phrase = w.getSeedPhrase()

    w.shieldWallet(random_password)
    seed = w.getEncryptedSeed()

    w_restored = Wallet.fromEncryptedSeedDict(
        seed, passphrase=random_password)

    assert w_restored.getSlatepackAddress() == slatepack_address
    assert w_restored.getSeedPhrase() == recovery_phrase


# this simply tests if runs, does not check any correctness
def test_decrypt():
    M = Mnemonic()
    master_seed_ref = M.entropyFromMnemonic(recovery_phrase)

    encrypted_seed = bytes.fromhex(seed['encrypted_seed'])
    nonce = bytes.fromhex(seed['nonce'])
    salt = bytes.fromhex(seed['salt'])

    w = Wallet(encrypted_seed=encrypted_seed, nonce=nonce, salt=salt)
    w.unshieldWallet(password, nonce=nonce, salt=salt)

    assert w.master_seed == master_seed_ref
    assert recovery_phrase == w.getSeedPhrase()

    w.shieldWallet(password, nonce=nonce, salt=salt)

    assert w.encrypted_seed == encrypted_seed


def test_invalid_password():
    invalid_password = 'Lt. Col. Frank Slade'
    try:
        Wallet.fromEncryptedSeedDict(
            seed, passphrase=invalid_password)
    except Exception as e:
        assert str(e) == 'MAC check failed'


def test_restore_from_encrypted_seed():
    w = Wallet.fromEncryptedSeedDict(
        seed, passphrase=password)

    M = Mnemonic()
    master_seed_ref = M.entropyFromMnemonic(recovery_phrase)
    assert w.master_seed == master_seed_ref

    address = w.getSlatepackAddress(path='m/0/1/0')
    assert address == 'grin14kgku7l5x6te3arast3p59zk4rteznq2ug6kmmypf2d6z8md76eqg3su35'


# @pytest.mark.skip(reason='wip')
def test_from_seed_phrase():
    w = Wallet.fromSeedPhrase(recovery_phrase)
    for label, path, expected_address in accounts:
        address = w.getSlatepackAddress(path=path)
        print()
        print(path)
        print(address)
        print(expected_address)
        assert address == expected_address


# @pytest.mark.skip(reason='wip')
def test_grin_plusplus():
    # seed_words = 'dehydrate opened lilac elapse subtly prying swept ruby liar veteran wife afloat strained camp tugs pager dual tomorrow aimless boxes saucepan invoke utensils vapidly lilac'
    # just_seed_words = ' '.join(seed_words.split(' ')[0:-1])
    # print(len(seed_words.split(' ')))
    # M = Mnemonic()
    # master_seed = M.entropyFromMnemonic(just_seed_words)

    # print('by', bindings.crypto_sign_SECRETKEYBYTES)
    seed = bytes.fromhex('f9a0e73d3cd533368f75ff63cbd97b2100beffbc339cdfa5c203c1a022d9cf11')
    # assert seed == master_seed

    # I AM VOLDEMORT
    m = hmac.new('IamVoldemort'.encode('utf8'), digestmod=sha512)
    m.update(seed)
    secret = m.digest()

    # derive the seed at the path
    bip32 = BIP32(chaincode=secret[32:], privkey=secret[:32])
    sk_der = bip32.get_privkey_from_path('m/0/1/0')

    # compute the blake2 hash of that key and that is ed25519 seed
    seed_blake = blake2b(sk_der, digest_size=32).digest()

    # get the ed25519 secret key and public key from it
    pk, sk = bindings.crypto_sign_seed_keypair(seed_blake)

    # compute the slatepack address
    slatepack_address = Bech32Encoder.Encode('grin', pk)

    assert slatepack_address == 'grin1q6qnz4y7chpmmjeazw59tcdhd5tejx00m96xdy5jkv30t4v6fnxqv9kwer'



import hashlib
import base64
import math

from io import BytesIO

from age.file import Decryptor as ageStreamDecrypt
from age.keys.agekey import AgePrivateKey

from pyrage import x25519 as age_x25519

from bip_utils import Bech32Encoder

from mimblewimble.serializer import Serializer
from mimblewimble.crypto.age import AgeMessage

from mimblewimble.helpers.encryption import ageX25519Encrypt, ageX25519Decrypt

from mimblewimble.models.slatepack.address import SlatepackAddress
from mimblewimble.models.slatepack.metadata import SlatepackVersion
from mimblewimble.models.slatepack.metadata import SlatepackMetadata

from mimblewimble.models.slatepack.message import EMode

raw = b'\xfb\xcf6\xfd\x01\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02{age-encryption.org/v1\n-> X25519 F/H3hf63Hrq2MH+O1naWm6+M8184u3S7oeHQ4VHiMzY\npbUcLLHdT5PGn/oYeiD+tsUO1oexurSZmSVu8Hcv7mc\n-> Es[-grease )vK1eC lVxuS`,6\nSkmgyCDG0wPb0wvGeYp6XbdSyMONfXYoRPLyQSlkse8tDxkTHKwjoz6Y3t2IIqHv\nU/dxyW2iZyZ1UVFRrlNAVbXbEBwqhpaFjDdRcivhbDo1uTL5CDNvJz89w+ivaoUc\n\n--- nq0hJv/bV3kLRrqKU9wlkexF2STZDmBf0hlees/HyWw\n\x8cA\xc3+\x1a\xaf\x10\x93T\x18\x0f\xf9J\x14\xaf\xb0\xcdw\xb9\x10\xa7\x90>V-\xc3=D\xafwc\xdeo\x08\xefx\xdd\x19\x85DtG\xad\xc8\x88*\x84\xe5Z\x07G\x95\x08\xad\xca\xb8\x81N\x1aJ\xfc\x97z\xa8\xfcR\xd8Y\xa4\x90\xcc[\x05\x989\xb1\x9d\xd1U\xdcsJu\xc7\xc4\xb9\x91\x03\x07\x08C\xd5\xfb>s\x80"\t\x85)\x8c(\x1dL\x9f\xa2\xea\xae\xf2f\xf8~\xfe\xc8\xc8\x14\xdeKD\xd8\x8b2\xd2\xca\xf4\xbdM\xcf\xe8_\x88\x03$=\xe1\x11(5g\xde\x9cF,S\xec0#*.\xa0\xe6H\x0f\x94\x8b\x02\xd1\xd9v\xea\xb8u%Dp\xb4\xb9\x88Jvz\xb9\x07\x1e$\x0f\xf5\x80\xe5\xbc\xe5I\x8b\x9f\xc8\xe1\x04\xec0T\x99\xf1cC\xd5\x90\xb27\x18\x1d\x88\x9d\x85\xea\x07\x11\xc4\x92\x0c\x07\xee\xc5\xb8\x1f\xf2&\xf29\x00R\xd4\xb6\xbfe\x18\xfd\xdd\x1a\x8f\xf0\xc3$\xfc0\xf3\xdc`\xed\xd0\xbd\xaa\xbd\t\x9d\xf1\x93.\xa6\xa6C\x0f\x0b;\x9er~P\xbd\xe1\xd5\xf5\x1f\xda>}\xe5J\x93\xc8\xde\xf4^\x9f\xbd\xbblK\xf3\xe3hP\xef\xdc\xb2\xbd\x05\xf1\x08\x9e\xe0\x96\xaeg\xd9\x1b\xc7\xf2/\xb8\xbe%aY\xdf\xc6T'
s = Serializer()
s.write(raw)


def test_age_message_serialization():
    serializer = s

    error_check_code = serializer.read(4)

    version = SlatepackVersion.deserialize(serializer)
    emode = EMode(int.from_bytes(serializer.read(1), 'big'))

    opt_flags = int.from_bytes(serializer.read(2), 'big')
    opt_fields_len = int.from_bytes(serializer.read(4), 'big')

    metadata = SlatepackMetadata()
    if opt_flags & 0x01 == 0x01:
        sender = SlatepackAddress.deserialize(serializer)
        metadata = SlatepackMetadata(sender=sender)

    payload_size = int.from_bytes(serializer.read(8), 'big')
    payload = serializer.read(payload_size)

    # checksum

    print()
    serializer.resetPointer()
    serializer.read(4)
    preimage = serializer.readremaining()
    hashed = hashlib.sha256(preimage).digest()
    hashed = hashlib.sha256(hashed).digest()
    checksum = hashed[0:4]
    if checksum == error_check_code:
        print('checksum correct')
    else:
        print('checksum mismatch')

    print('emode is', emode)
    if emode == EMode.ENCRYPTED:
        print('is encrypted')
    else:
        print('is not encrypted')

    # m/0/1/0 alice x25519 secret key
    x25519sk = bytes.fromhex('90ad4d97dd2f7dd5360375f2ad72011489baa8beb84f1109b2bdb79b4183476a')

    ciphertext = payload

    age_secret_key = Bech32Encoder.Encode(
        'age-secret-key-', x25519sk)
    decrypted = ageX25519Decrypt(ciphertext, age_secret_key)

    print('decrypted payload')
    print(decrypted)

    d = Serializer()
    d.write(decrypted)

    size = int.from_bytes(d.read(4), 'big')

    inner_buffer = Serializer()
    inner_buffer.write(d.read(size))

    opt_flags = int.from_bytes(inner_buffer.read(2), 'big')

    sender = None
    if opt_flags & 0x01 == 0x01:
        sender = SlatepackAddress.deserialize(inner_buffer)

    recipients = []
    if opt_flags & 0x02 == 0x02:
        num_recipients = int.from_bytes(inner_buffer.read(2), 'big')
        for i in range(num_recipients):
            recipient = SlatepackAddress.deserialize(inner_buffer)
            recipients.append(recipient)

    print('sender', sender.toBech32())
    print('recipients', recipients)


    '''
    serializer = Serializer()

    age_message = AgeMessage.deserialize(s)

    serializer = Serializer()
    age_message.serialize(serializer)

    assert serializer.readall() == raw

    print()
    for recipient in age_message.header.recipients:
        print(recipient._type.decode(), [arg.decode() for arg in recipient.args])
    '''


def test_age_recipient_stanza():
    path = 'm/0/1/0'
    slatepack_address = 'grin14kgku7l5x6te3arast3p59zk4rteznq2ug6kmmypf2d6z8md76eqg3su35'
    ed25519pk = 'ad916e7bf4369798f47d82e21a1456a8d7914c0ae2356dec814a9ba11f6df6b2'
    ed25519sk = '7f7394cd934ba03a26d5774ef45b5969670d0ce1e4bbc1174e6a871f507af0ddad916e7bf4369798f47d82e21a1456a8d7914c0ae2356dec814a9ba11f6df6b2'
    x25519pk = '585b020389181ac540a9e268eb5f5d7d86297bad2ee0c8fb5eb533f92ce66d4d'
    x25519sk = '90ad4d97dd2f7dd5360375f2ad72011489baa8beb84f1109b2bdb79b4183476a'
    agepk = 'age1tpdsyqufrqdv2s9fuf5wkh6a0krzj7ad9msv3767k5eljt8xd4xscfzxhm'
    agesk = 'AGE-SECRET-KEY-1JZK5M97A9A7A2DSRWHE26USPZJYM4297HP83ZZDJHKMEKSVRGA4QG7GS5A'

    # header data
    ephemeral_share = b'F/H3hf63Hrq2MH+O1naWm6+M8184u3S7oeHQ4VHiMzY' + b'=='
    encrypted_file_key = b'pbUcLLHdT5PGn/oYeiD+tsUO1oexurSZmSVu8Hcv7mc' + b'=='
    header_mac = b'nq0hJv/bV3kLRrqKU9wlkexF2STZDmBf0hlees/HyWw' + b'=='

    # payload data
    encrypted_payload = b'\x8cA\xc3+\x1a\xaf\x10\x93T\x18\x0f\xf9J\x14\xaf\xb0\xcdw\xb9\x10\xa7\x90>V-\xc3=D\xafwc\xdeo\x08\xefx\xdd\x19\x85DtG\xad\xc8\x88*\x84\xe5Z\x07G\x95\x08\xad\xca\xb8\x81N\x1aJ\xfc\x97z\xa8\xfcR\xd8Y\xa4\x90\xcc[\x05\x989\xb1\x9d\xd1U\xdcsJu\xc7\xc4\xb9\x91\x03\x07\x08C\xd5\xfb>s\x80"\t\x85)\x8c(\x1dL\x9f\xa2\xea\xae\xf2f\xf8~\xfe\xc8\xc8\x14\xdeKD\xd8\x8b2\xd2\xca\xf4\xbdM\xcf\xe8_\x88\x03$=\xe1\x11(5g\xde\x9cF,S\xec0#*.\xa0\xe6H\x0f\x94\x8b\x02\xd1\xd9v\xea\xb8u%Dp\xb4\xb9\x88Jvz\xb9\x07\x1e$\x0f\xf5\x80\xe5\xbc\xe5I\x8b\x9f\xc8\xe1\x04\xec0T\x99\xf1cC\xd5\x90\xb27\x18\x1d\x88\x9d\x85\xea\x07\x11\xc4\x92\x0c\x07\xee\xc5\xb8\x1f\xf2&\xf29\x00R\xd4\xb6\xbfe\x18\xfd\xdd\x1a\x8f\xf0\xc3$\xfc0\xf3\xdc`\xed\xd0\xbd\xaa\xbd\t\x9d\xf1\x93.\xa6\xa6C\x0f\x0b;\x9er~P\xbd\xe1\xd5\xf5\x1f\xda>}\xe5J\x93\xc8\xde\xf4^\x9f\xbd\xbblK\xf3\xe3hP\xef\xdc\xb2\xbd\x05\xf1\x08\x9e\xe0\x96\xaeg\xd9\x1b\xc7\xf2/\xb8\xbe%aY\xdf\xc6T'

    # decode data from the header
    decoded_ephemeral_share = base64.b64decode(
        ephemeral_share)
    assert decoded_ephemeral_share.hex() == '17f1f785feb71ebab6307f8ed676969baf8cf35f38bb74bba1e1d0e151e23336'


    decoded_encrypted_file_key = base64.b64decode(
        encrypted_file_key)
    assert decoded_encrypted_file_key.hex() == 'a5b51c2cb1dd4f93c69ffa187a20feb6c50ed687b1bab49999256ef0772fee67'

    decoded_header_mac = base64.b64decode(
        header_mac)
    assert decoded_header_mac.hex() == '9ead2126ffdb57790b46ba8a53dc2591ec45d924d90e605fd2195e7acfc7c96c'

    public_key = bytes.fromhex(x25519pk)
    private_key = bytes.fromhex(x25519sk)

    salt = decoded_ephemeral_share + public_key

    from nacl.bindings import crypto_scalarmult

    key_material = crypto_scalarmult(
        private_key, decoded_ephemeral_share)

    assert key_material.hex() == '5ab5690070158a90a7dfefa4639381ff676782cd6fce9c2d50d31f2fed389c77'

    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.kdf.hkdf import HKDF
    from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305

    AGE_X25519_HKDF_LABEL = b'age-encryption.org/v1/X25519'

    key = HKDF(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        info=AGE_X25519_HKDF_LABEL,
        backend=default_backend()
    ).derive(key_material)

    assert key.hex() == 'd008d262cc77afdcf96c0a0363ec9894e1bc840a218a5afb86eacb1c5168ba0a'

    ZERO_NONCE = b'\00' * 12

    decrypted_file_key = ChaCha20Poly1305(key).decrypt(
        ZERO_NONCE,
        decoded_encrypted_file_key,
        None)

    # print('decrypted_file_key', decrypted_file_key.hex())

    assert decrypted_file_key.hex() == 'd236c4a664b6e1d909a06bb8b7cdbd47'
    assert len(decrypted_file_key) == 16

    # assert decrypted_file_key == bytes.fromhex('d008d262cc77afdcf96c0a0363ec9894e1bc840a218a5afb86eacb1c5168ba0a')


    from cryptography.hazmat.primitives.hmac import HMAC

    HEADER_HKDF_LABEL = b'header'
    salt = b''
    hkdfval = HKDF(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        info=HEADER_HKDF_LABEL,
        backend=default_backend()
    ).derive(decrypted_file_key)
    assert hkdfval.hex() == 'e9e4ff52ce1c1c2f7c8d280fc6f869ade9bc005b28c07a36c17ee8839e59527f'

    hmac = HMAC(
        key=hkdfval,
        algorithm=hashes.SHA256(),
        backend=default_backend()
    )

    mac_message = b'age-encryption.org/v1\n-> X25519 F/H3hf63Hrq2MH+O1naWm6+M8184u3S7oeHQ4VHiMzY\npbUcLLHdT5PGn/oYeiD+tsUO1oexurSZmSVu8Hcv7mc\n-> Es[-grease )vK1eC lVxuS`,6\nSkmgyCDG0wPb0wvGeYp6XbdSyMONfXYoRPLyQSlkse8tDxkTHKwjoz6Y3t2IIqHv\nU/dxyW2iZyZ1UVFRrlNAVbXbEBwqhpaFjDdRcivhbDo1uTL5CDNvJz89w+ivaoUc\n---'
    hmac.update(mac_message)

    # THIS FAILS TODO
    # hmac.verify(decoded_header_mac)

    # proceed to decrypt and prove decrypted payload is correct
    stream = BytesIO()
    stream.write(encrypted_payload)

    stream.seek(0)
    nonce = stream.read(16)
    assert nonce.hex() == '8c41c32b1aaf109354180ff94a14afb0'
    PAYLOAD_HKDF_LABEL = b'payload'
    stream_key = HKDF(
        algorithm=hashes.SHA256(),
        length=32,
        salt=nonce,
        info=PAYLOAD_HKDF_LABEL,
        backend=default_backend()
    ).derive(decrypted_file_key)

    assert stream_key.hex() == '8d1229aaaf0cdf5e7917ab2a725ff68f3338d76bd2734bc9e63fde486e12cdf9'
    assert len(stream_key) == 32

    ciphertext = stream.read()
    assert ciphertext == b'\xcdw\xb9\x10\xa7\x90>V-\xc3=D\xafwc\xdeo\x08\xefx\xdd\x19\x85DtG\xad\xc8\x88*\x84\xe5Z\x07G\x95\x08\xad\xca\xb8\x81N\x1aJ\xfc\x97z\xa8\xfcR\xd8Y\xa4\x90\xcc[\x05\x989\xb1\x9d\xd1U\xdcsJu\xc7\xc4\xb9\x91\x03\x07\x08C\xd5\xfb>s\x80"\t\x85)\x8c(\x1dL\x9f\xa2\xea\xae\xf2f\xf8~\xfe\xc8\xc8\x14\xdeKD\xd8\x8b2\xd2\xca\xf4\xbdM\xcf\xe8_\x88\x03$=\xe1\x11(5g\xde\x9cF,S\xec0#*.\xa0\xe6H\x0f\x94\x8b\x02\xd1\xd9v\xea\xb8u%Dp\xb4\xb9\x88Jvz\xb9\x07\x1e$\x0f\xf5\x80\xe5\xbc\xe5I\x8b\x9f\xc8\xe1\x04\xec0T\x99\xf1cC\xd5\x90\xb27\x18\x1d\x88\x9d\x85\xea\x07\x11\xc4\x92\x0c\x07\xee\xc5\xb8\x1f\xf2&\xf29\x00R\xd4\xb6\xbfe\x18\xfd\xdd\x1a\x8f\xf0\xc3$\xfc0\xf3\xdc`\xed\xd0\xbd\xaa\xbd\t\x9d\xf1\x93.\xa6\xa6C\x0f\x0b;\x9er~P\xbd\xe1\xd5\xf5\x1f\xda>}\xe5J\x93\xc8\xde\xf4^\x9f\xbd\xbblK\xf3\xe3hP\xef\xdc\xb2\xbd\x05\xf1\x08\x9e\xe0\x96\xaeg\xd9\x1b\xc7\xf2/\xb8\xbe%aY\xdf\xc6T'
    plaintext = b''

    PLAINTEXT_BLOCK_SIZE = 64 * 1024
    CIPHERTEXT_BLOCK_SIZE = PLAINTEXT_BLOCK_SIZE + 16

    aead = ChaCha20Poly1305(stream_key)
    blocks = math.ceil(len(ciphertext) / CIPHERTEXT_BLOCK_SIZE)

    def _chunk(data, size):
        for i in range(0, len(data), size):
            yield data[i : i + size]


    for nonce, block in enumerate(_chunk(ciphertext, CIPHERTEXT_BLOCK_SIZE)):
        last_block = nonce == blocks - 1
        packed_nonce = nonce.to_bytes(
            11,
            byteorder='big',
            signed=False
        ) + (b'\x01' if last_block else b'\x00')

        plaintext += aead.decrypt(
            nonce=packed_nonce,
            data=block,
            associated_data=None)

    assert plaintext == b'\x00\x00\x00B\x00\x01?grin1m4krnajw792zxfyldu79jssh0d3kzjtwpdn2wy7fysrfw4ej0waskurq76\x00\x04\x00\x03S\xfe\x05\xcd\x07\x8bK\x8f\x96\x0bV\xafo\x1a3\xb1\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x00\x00\x00\x00:\xdc\r\xe0\x00\x00\x00\x00\x00\xbe\xbc \x01\x00\x02S3\xf6~\x84^\xdbV\xe3\t\xd6\x9c\x1b\xc5k\x89T\xde\t\x82\xba,\xe4!\x13\x0br\x15\x19\xa7\xa7\x9a\x022\xd9\xb4\x99\x04=\xfb5\xce\x97R;9\xa6\r\x1b"\xca.\x972bpKu\xb3\xd6\x1e\x05\xd4\x96\xe9\x02\xddl9\xf6N\xf1T#$\x9fo<YB\x17{caIn\x0bf\xa7\x13\xc9$\x06\x97W2{\xbb\xad\x91n{\xf46\x97\x98\xf4}\x82\xe2\x1a\x14V\xa8\xd7\x91L\n\xe25m\xec\x81J\x9b\xa1\x1fm\xf6\xb2\x00'



    # TODO now we need to prove that the decrypted payload is correct

def test_pyrage_encryption():
    # Generate key pair
    identity = age_x25519.Identity.generate()
    print(str(identity))
    public_key = identity.to_public()
    print(public_key)
    plaintext = b"Hello, this is a test message!"

    # Test encryption and decryption with public key string
    recipients = [str(public_key)]
    ciphertext = ageX25519Encrypt(plaintext, recipients)
    decrypted = ageX25519Decrypt(ciphertext, str(identity))
    assert decrypted == plaintext, "Decryption failed for string public key"

    # Test empty plaintext
    plaintext = b""
    ciphertext = ageX25519Encrypt(plaintext, recipients)
    decrypted = ageX25519Decrypt(ciphertext, str(identity))
    assert decrypted == plaintext, "Decryption failed for empty plaintext"

def test_pyrage_case():
    secret_key = 'AGE-SECRET-KEY-16XFRV8AFSREQ6JGZ6JK3JW4X0DRVS9ENUPWRZDDEQ6024LG0GFTQ45M3J4'
    public_key = 'age1ax2qd3283yh0hr0d7zw04qjk7s2782qwgmpu7d9yju5rwqjymehsz8wjce'
    plaintext = b"Hello, this is a test message!"

    # Test encryption and decryption with public key string
    recipients = [str(public_key)]
    ciphertext = ageX25519Encrypt(plaintext, recipients)
    decrypted = ageX25519Decrypt(ciphertext, secret_key)
    assert decrypted == plaintext, "Decryption failed for string public key"

def test_age_both_ways():
    plaintext = b'This is a test message for age encryption and decryption.'
    age_secret_key = 'AGE-SECRET-KEY-1WZWET7A3CSR8YM446MGH2U3FNHTJ6G05856GL5X52X3WNTR9NA7S8W0C5H'
    recipients = ['age1f2yv3z4hx6rfdmect7uecjwyzfyuscrtnet6v0eydypugy2f93gsuwac7u']

    ciphertext = ageX25519Encrypt(plaintext, recipients)
    decrypted = ageX25519Decrypt(ciphertext, age_secret_key)

    assert decrypted == plaintext

def test_case_age_slatepacks():
    from mimblewimble.codecs import Base64Codec
    plaintext_b64 = 'AAAAAQSoF2K4xG0yKGSeKkvuXvUBTgZuwD6dhM/VUX04TUgqYNFxwotIqy+aeXupx8Mj5aoGAAAABvwjrAAAAAAAEzFUgAEAAhBwmMJIBoHTujsWq5xpbKkuWrCt+Q8Se2NTsTmOvK/7A9E/jeAC1nYNZR/NIp4hQfKmX+ExI71rXYigtUSJiWUjAwACAAEIW6FwSC3S89EdFe/CmR/XAYM5fMsIq14lWNv2V8VctWEBAAk9SY7XjQ8ntIT2UbZhxVnzXGVAdbxhx5d8GvlIxiAEDQAAAAAAAAKjLOo97MRxVi/LR/5eME/JzZaStaUi7OAdiKdHelMING3i8++Srdtlr0u0ZBFqzTokiVJx+mWEKkTThpRjFvGg8QoFybLu1tZF/2X9y3U7hZP47rEM5R0ui/83meyemkmhKfaRlruIsE1EHmrB9zETsFtO5WkJnmjiBkba/tANHx24Wu3LQwEzIQv8ilfAANy8eOR+s+unDEwfsIWFsKcfQ7rOCc4CWDG6507rtJhnLLcvbeMchPBFqnvGolVXnDN7/FVj6oF3ChQ7gojYSdRO06Pl9zaWMo2fu+kmTIoWNp+iNdpRZO9NteHJDkW2W6jFzOVnvfLWbhaERGBxJvFEyvYojydVGUJoVwMJ1x1XH6UqHC2uBBe8Fe9Eg4zcgfTIkEdl5V14B4L5qrIT9veK1sNL01V7ckyNqOJaA2YliPU0toa1LSauK7mG/AtUixqXbKGYLSetLOOXBOx4NGl/gbhpAK7WCmNkuUtc9yxhdADZJDyC4F/KylaYzxxkA3pbpf84BqkyrQOr7V5Z6pg+2VK3qrNVF06jf8w4tCMJbxtEEPlZy3TLKZ6z3ulzP6TGpZcLCzK/XkqJFIxv2M/he+XHrtB3wUK0VzMbs2S9d9agoSAMBF8GFrT80b2gH1SXc+ifPH3xBjmIbiGOcS91KW3HKrJdW4pUy3qhlpjWgc1J0OA9Ap7TYsuxa09coyZxie16I+KHrJe9+Ld3CMB2GwRa/vT85T2ag4OpYGz3/WG+0UWoDiUe1klPmdTVBaYGQJ4ZlvPtZ64AfXaBKdvx3ScqoUdvji0dHWoIv9RxSiYFHlgWsp9Y6+vgX2jm5ByIF6q+oY0t8E5KH21seItUBqBhstCtcMQuU3NfjeTlkW3LU17QMiWGJMrM+jFViYOMQ0jhUR9dY8DaceaQ2Jg3XO6ibTuo0WYR/Wf6OHytnekKizQDULPYhfmI/Tr/Vk4c8XpwiBSY8h7C4DONF3xceU+13AA='
    plaintext = Base64Codec.decode(plaintext_b64)
    # plaintext = b'This is a test message for age encryption and decryption.'
    age_secret_key = 'AGE-SECRET-KEY-1QPRVDN2X7RZ03AZUP3MQL5TWXVMWPVJQNVSZAQ58XHDJE6CL64HSW2XHVP'
    recipients = ['age1mrdqwkrdkj4hldcv7q94evatqe5reedc045vddlwd3742verj3cqfcxhr5']

    ciphertext = ageX25519Encrypt(plaintext, recipients)
    decrypted = ageX25519Decrypt(ciphertext, age_secret_key)

    assert decrypted == plaintext


from mimblewimble.models.transaction import TransactionKernel
from mimblewimble.models.transaction import TransactionOutput

from mimblewimble.entity import OutputDataEntity

from mimblewimble.wallet import Wallet


def test_coinbase_construction():
    w = Wallet.initialize()

    amount = 60
    kernel, output = w.createCoinbase(amount, path='m/0/1/0')

    assert isinstance(kernel, TransactionKernel)
    assert isinstance(output, OutputDataEntity)
    assert isinstance(output.output, TransactionOutput)
    assert output.path == 'm/0/1/0'


def test_srs_flow():
    alice_path = 'm/0/1/0'
    alice_wallet = Wallet.initialize()
    alice_address = alice_wallet.getSlatepackAddress()

    bob_path = 'm/0/1/0'
    bob_wallet = Wallet.initialize()
    bob_address = bob_wallet.getSlatepackAddress()

    coinbase_amount = 60000000000
    kernel, output = alice_wallet.createCoinbase(
        coinbase_amount, path=alice_path)

    num_change_outputs = 1
    amount = 30000000000
    fee_base = 7000000
    block_height = 99999

    # send
    send_slate, alice_secret_key, alice_secret_nonce, change_outputs = alice_wallet.send(
        [output], num_change_outputs, amount, fee_base, block_height,
        path=alice_path, receiver_address=bob_address)

    # receive
    receive_slate, bob_secret_key, bob_secret_nonce = bob_wallet.receive(send_slate)

    # finalize
    finalized_slate = alice_wallet.finalize(
        receive_slate, path=alice_path)



import unittest

from mimblewimble.blockchain import ProofOfWork, BlockHeader


class BlockHeaderTest(unittest.TestCase):
    def test_deserialize(self):
        version = 1
        height = 2
        timestamp = 3

        previousBlockHash = bytes.fromhex('0102030405060708090A0B0C0D0E0F1112131415161718191A1B1C1D1E1F2021')
        previousRoot = bytes.fromhex('1102030405060708090A0B0C0D0E0F1112131415161718191A1B1C1D1E1F2021')
        outputRoot = bytes.fromhex('2102030405060708090A0B0C0D0E0F1112131415161718191A1B1C1D1E1F2021')
        rangeProofRoot = bytes.fromhex('3102030405060708090A0B0C0D0E0F1112131415161718191A1B1C1D1E1F2021')
        kernelRoot = bytes.fromhex('4102030405060708090A0B0C0D0E0F1112131415161718191A1B1C1D1E1F2021')
        totalKernelOffset = bytes.fromhex('5102030405060708090A0B0C0D0E0F1112131415161718191A1B1C1D1E1F2021')

        outputMMRSize = 4
        kernelMMRSize = 7
        totalDifficulty = 1000
        scalingDifficulty = 10
        nonce = 123456

        proofOfWork = ProofOfWork(bytes.fromhex('00001010'), [])

        block_header = BlockHeader(
            version,
            height,
            timestamp,
            previousBlockHash,
            previousRoot,
            outputRoot,
            rangeProofRoot,
            kernelRoot,
            totalKernelOffset,
            outputMMRSize,
            kernelMMRSize,
            totalDifficulty,
            scalingDifficulty,
            nonce,
            proofOfWork)

        assert block_header.getNumOutputs() == 3
        assert block_header.getNumKernels() == 4



import hashlib
import json # TODO remove after debugging
from io import BytesIO

from mimblewimble.mmr.index import MMRIndex
from mimblewimble.serializer import Serializer

from mimblewimble.consensus import Consensus
from mimblewimble.models.short_id import ShortId

from mimblewimble.models.transaction import TransactionInput
from mimblewimble.models.transaction import TransactionOutput
from mimblewimble.models.transaction import TransactionBody
from mimblewimble.models.transaction import BlindingFactor


class ProofOfWork:
    def __init__(self, edgeBits, proofNonces):
        self.edgeBits = edgeBits # 1 byte
        self.proofNonces = proofNonces

    def getEdgeBits(self):
        return self.edgeBits

    def getProofNonces(self):
        return self.proofNonces

    def isPrimary(self):
        return Consensus.isPrimary(self.edgeBits)

    def isSecondary(self):
        return Consensus.isSecondary(self.edgeBits)

    def serialize(self, serializer):
        serializer.write(self.getEdgeBits().to_bytes(1, 'big'))
        serializer.write(self.serializeCycle())

    def serializeCycle(self):
        bytes_len = int(((self.getEdgeBits()*Consensus.proofsize)+7)/8)
        serialized_bytes = bytearray(bytes_len)
        for n in range(len(self.getProofNonces())):
            for bit in range(int(self.getEdgeBits())):
                nonce = self.proofNonces[n]
                if nonce & (1 << bit) != 0:
                    positionTemp = (n*self.edgeBits)+bit
                    p = int(positionTemp/8)
                    serialized_bytes[p] |= (1 << (positionTemp % 8))
        return serialized_bytes

    @classmethod
    def deserialize(self, B):
        edgeBits = int.from_bytes(B.read(1), 'big')
        bytes_len = int(((edgeBits*Consensus.proofsize)+7)/8)
        bits = B.read(bytes_len)
        proofNonces = self.deserializeProofNonces(bits, edgeBits)
        return ProofOfWork(edgeBits, proofNonces)

    def deserializeProofNonces(bits, edgeBits):
        if edgeBits == 0 or edgeBits > 63:
            raise ValueError('Invalid number of edge bits {0}'.format(str(edgeBits)))
        uint8_t1 = b'\x00\x00\x00\x00\x00\x00\x00\x01'
        proofNonces = []
        for n in range(Consensus.proofsize):
            proofNonce = 0
            for bit in range(edgeBits):
                positionTemp = (n*edgeBits)+bit
                p = int(positionTemp/8)
                if int(bits[p]) & (1 << (positionTemp % 8)):
                    proofNonce |= 1 << bit
            proofNonces.append(proofNonce)
        return proofNonces

    def getHash(self):
        cycle = self.serializeCycle()
        return hashlib.blake2b(cycle, digest_size=32).digest()


class BlockHeader:
    def __init__(self,
                 version,
                 height,
                 timestamp,
                 previousBlockHash,
                 previousRoot,
                 outputRoot,
                 rangeProofRoot,
                 kernelRoot,
                 totalKernelOffset,
                 outputMMRSize,
                 kernelMMRSize,
                 totalDifficulty,
                 scalingDifficulty,
                 nonce,
                 proofOfWork):
        self.version = version
        self.height = height
        self.timestamp = timestamp
        self.previousBlockHash = previousBlockHash
        self.previousRoot = previousRoot
        self.outputRoot = outputRoot
        self.rangeProofRoot = rangeProofRoot
        self.kernelRoot = kernelRoot
        self.totalKernelOffset = totalKernelOffset
        self.outputMMRSize = outputMMRSize
        self.kernelMMRSize = kernelMMRSize
        self.totalDifficulty = totalDifficulty
        self.scalingDifficulty = scalingDifficulty
        self.nonce = nonce
        self.proofOfWork = proofOfWork

    # getters

    def getVersion(self):
        return self.version

    def getHeight(self):
        return self.height

    def getPreviousHash(self):
        return self.previousBlockHash

    def getPreviousRoot(self):
        return self.previousRoot

    def getTimestamp(self):
        return self.timestamp

    def getTotalDifficulty(self):
        return self.totalDifficulty

    def getScalingDifficulty(self):
        return self.scalingDifficulty

    def getTotalScalingDifficulty(self):
        return self.scalingDifficulty

    def getTotalKernelOffset(self):
        return self.totalKernelOffset

    def getNonce(self):
        return self.nonce

    # pow

    def getProofOfWork(self):
        return self.proofOfWork

    def getEdgeBits(self):
        return self.proofOfWork.getEdgeBits()

    def getProofNonces(self):
        return self.proofOfWork.getProofNonces()

    def isPrimaryPoW(self):
        return self.proofOfWork.isPrimary()

    def isSecondaryPoW(self):
        return self.proofOfWork.isSecondary()

    # Merklish root stuffz

    def getOutputRoot(self):
        return self.outputRoot

    def getRangeProofRoot(self):
        return self.rangeProofRoot

    def getKernelRoot(self):
        return self.kernelRoot

    # Merkle Mountain Range Sizes

    def getOutputMMRSize(self):
        return self.outputMMRSize

    def getKernelMMRSize(self):
        return self.kernelMMRSize

    def getNumOutputs(self):
        return MMRIndex.at(self.outputMMRSize).getLeafIndex()

    def getNumKernels(self):
        return MMRIndex.at(self.kernelMMRSize).getLeafIndex()

    # serialization / deserialization

    def serialize(self, serializer: Serializer):
        serializer.write(self.version.to_bytes(2, 'big'))
        serializer.write(self.height.to_bytes(8, 'big'))
        serializer.write(self.timestamp.to_bytes(8, 'big'))
        serializer.write(self.previousBlockHash)
        serializer.write(self.previousRoot)
        serializer.write(self.outputRoot)
        serializer.write(self.rangeProofRoot)
        serializer.write(self.kernelRoot)
        serializer.write(self.totalKernelOffset.serialize()) # blinding factor
        serializer.write(self.outputMMRSize.to_bytes(8, 'big'))
        serializer.write(self.kernelMMRSize.to_bytes(8, 'big'))
        serializer.write(self.totalDifficulty.to_bytes(8, 'big'))
        serializer.write(self.scalingDifficulty.to_bytes(4, 'big'))
        serializer.write(self.nonce.to_bytes(8, 'big'))
        self.proofOfWork.serialize(serializer)

    @classmethod
    def deserialize(self, B: Serializer):
        version = int.from_bytes(B.read(2), 'big')
        height = int.from_bytes(B.read(8), 'big')
        timestamp = int.from_bytes(B.read(8), 'big')
        previousBlockHash = B.read(32)
        previousRoot = B.read(32)
        outputRoot = B.read(32)
        rangeProofRoot = B.read(32)
        kernelRoot = B.read(32)

        totalKernelOffset = BlindingFactor.deserialize(B.read(32))

        outputMMRSize = int.from_bytes(B.read(8), 'big')
        kernelMMRSize = int.from_bytes(B.read(8), 'big')
        totalDifficulty = int.from_bytes(B.read(8), 'big')
        scalingDifficulty = int.from_bytes(B.read(4), 'big')
        nonce = int.from_bytes(B.read(8), 'big')

        proofOfWork = ProofOfWork.deserialize(B)

        return BlockHeader(version,
                           height,
                           timestamp,
                           previousBlockHash,
                           previousRoot,
                           outputRoot,
                           rangeProofRoot,
                           kernelRoot,
                           totalKernelOffset,
                           outputMMRSize,
                           kernelMMRSize,
                           totalDifficulty,
                           scalingDifficulty,
                           nonce,
                           proofOfWork)


    def toJSON(self):
        cuckooSolution = b''
        for proofNonce in self.getProofOfWork().getProofNonces():
            cuckooSolution += proofNonce.to_bytes(8, 'big')

        return {
            'height': self.getHeight(),
            'hash': self.getHash().hex(),
            'version': self.getVersion(),
            'timestamp_raw': self.getTimestamp(),
            'timestamp_local': self.getTimestamp(), # TODO convert to local
            'timestamp': self.getTimestamp(), # TODO convert to UTC
            'previous': self.getPreviousHash().hex(),
            'prev_root': self.getPreviousRoot().hex(),
            'kernel_root': self.getKernelRoot().hex(),
            'output_root': self.getOutputRoot().hex(),
            'range_proof_root': self.getRangeProofRoot().hex(),
            'output_mmr_size': self.getOutputMMRSize(),
            'kernel_mmr_size': self.getKernelMMRSize(),
            'total_kernel_offset': self.getTotalKernelOffset().hex(),
            'secondary_scaling': self.getScalingDifficulty(),
            'total_difficulty': self.getTotalDifficulty(),
            'nonce': self.getNonce(),
            'edge_bits': self.getProofOfWork().getEdgeBits(),
            'cuckoo_solution': cuckooSolution.hex()
        }

    @classmethod
    def fromDict(self, O: dict):
        return BlockHeader(O['version'],
                           O['height'],
                           O['timestamp'],
                           O['previousBlockHash'],
                           O['previousRoot'],
                           O['rangeProofRoot'],
                           O['kernelRoot'],
                           O['totalKernelOffset'],
                           O['outputMMRSize'],
                           O['kernelMMRSize'],
                           O['totalDifficulty'],
                           O['scalingDifficulty'],
                           O['nonce'],
                           ProofOfWork.deserialize(O['proofOfWork']))

    @classmethod
    def fromJSON(self, jsonString: str):
        O = json.loads(jsonString)
        return BlockHeader.fromDict(O)

    def getPreProofOfWork(self, serializer):
        serializer = BytesIO()
        self.serialize(serializer)
        return serializer.getvalue()

    # hashing

    def getHash(self):
        return self.proofOfWork.getHash()

    def shortHash(self):
        # TODO
        pass


class FullBlock:
    def __init__(self, header: BlockHeader, body: TransactionBody, validated=False):
        self.header = header
        self.body = body
        self.validated = validated

    def getHeader(self):
        return self.header

    def getTransactionBody(self):
        return self.body

    def getInputs(self):
        return self.body.getInputs()

    def getOutputs(self):
        return self.body.getOutputs()

    def getKernels(self):
        return self.body.getKernels()

    def getInputCommitments(self):
        return [_input.getCommitment() for _input in self.getInputs()]

    def getOutputCommitments(self):
        return [_output.getCommitment() for _output in self.getOutputs()]

    def getTotalFees(self):
        return self.body.calcFee()

    def calcWeight(self):
        return self.body.calcWeight(self.getHeight())

    def getHeight(self):
        return self.header.getHeight()

    def getPreviousHash(self):
        return self.header.getPreviousHash()

    def getTotalDifficulty(self):
        return self.header.getTotalDifficulty()

    def getTotalKernelOffset(self):
        return self.header.getTotalKernelOffset()

    def serialize(self):
        serializer = Serializer()
        self.header.serialize(serializer)
        self.body.serialize(serializer)
        return serializer.getvalue()

    @classmethod
    def deserialize(self, serializer: Serializer):
        header = BlockHeader.deserialize(serializer)
        body = TransactionBody.deserialize(serializer)
        return FullBlock(header, body)

    def toJSON(self):
        # transaction outputs
        outputs = []
        for _output in self.getOutputs():
            output_json = _output.toJSON()
            output_json['block_height'] = self.getHeight()
            outputs.append(output_json)

        return {
            'header': self.header.toJSON(),
            'inputs': [_input.toJSON() for _input in self.getInputs()],
            'outputs': outputs,
            'kernels': [kernel.toJSON() for kernel in self.getKernels()]
        }

    def getHash(self):
        return self.header.getHash()

    def wasValidated(self):
        return self.validated

    def markAsValidated(self):
        self.validated = True


class CompactBlock:
    def __init__(self, header, nonce, fullOutputs, fullKernels, shortIds):
        self.header = header
        self.nonce = nonce
        self.outputs = fullOutputs
        self.kernels = fullKernels
        self.short_ids = shoftIds

    # getters

    def getHeader(self):
        return self.header

    def getNonce(self):
        return self.nonce

    def getOutputs(self):
        return self.outputs

    def getKernels(self):
        return self.kernels

    def getShortIds(self):
        return self.short_ids

    def getPreviousHash(self):
        return self.header.getPreviousHash()

    def getHeight(self):
        return self.header.getHeight()

    def getTotalDifficulty(self):
        return self.header.getTotalDifficulty()

    # serialization / deserialization

    def serialize():
        # TODO check if there is a need to set byteorder='big'
        bytes_nonce = self.nonce.to_bytes(64)

        bytes_num_outputs = len(self.getOutputs()).to_bytes(64)
        bytes_num_kernels = len(self.getKernels()).to_bytes(64)
        bytes_num_short_ids = len(self.getShortIds()).to_bytes(64)

        outputs = self.getOutputs()
        bytes_outputs = outputs[0].serialize()
        for _output in outputs[1:]:
            bytes_outputs += _output.serialize()

        kernels = self.getKernels()
        bytes_kernels = kernels[0].serialize()
        for _kernel in kernels[1:]:
            bytes_kernels += _kernel.serialize()

        short_ids = self.getShortIds()
        bytes_short_ids = short_ids[0].serialize()
        for _short_id in short_ids[1:]:
            bytes_short_ids += _short_id.serialize()

        return bytes_nonce + bytes_num_otputs + bytes_num_kernels + bytes_num_short_ids + bytes_outputs + bytes_kernels + bytes_short_ids

    @classmethod
    def deserialize(byteString: bytes):
        nonce = int(byteString[0:64])

        numOutputs = int(byteString[65:128])
        numKernels = int(byteString[129:192])
        numShortIds = int(byteString[193:256])

        # TODO byteString should be shifted accordingly
        # TODO check how much it should be shifted at each iteration
        outputs = []
        for i in range(numOutputs):
            outputs.prepend(TransactionOutput.deserialize(byteString))

        kernels = []
        for i in range(numKernels):
            kernels.prepend(TransactionKernel.deserialize(byteString))

        shortIds = []
        for i in range(numShortIds):
            shortIds.prepend(ShortId.deserialize(byteString))

        return CompactBlock(nonce, outputs, kernels, shortIds)

    def toJSON():
        # transaction outputs
        outputs = []
        for _output in self.getOutputs():
            output_json = _output.toJSON()
            output_json['block_height'] = self.getHeight()
            outputs.append(output_json)

        return {
            'header': self.header.toJSON(),
            'inputs': [_input.toJSON() for _input in self.getInputs()],
            'outputs': outputs,
            'kernels': [kernel.toJSON() for kernel in self.getKernels()]
        }

    # hashing
    def __hash__():
        return hash(self.header)




import unittest
import os

from mimblewimble.mnemonic import Mnemonic

# https://github.com/mimblewimble/grin/blob/master/keychain/src/mnemonic.rs#L184
dataset = [
    (
        '00000000000000000000000000000000',
        'abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about',
        'c55257c360c07c72029aebc1b53c05ed0362ada38ead3e3e9efa3708e53495531f09a6987599d18264c1e1c92f2cf141630c7a3c4ab7c81b2f001698e7463b04'),
    (
        '7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f',
        'legal winner thank year wave sausage worth useful legal winner thank yellow',
        '2e8905819b8723fe2c1d161860e5ee1830318dbf49a83bd451cfb8440c28bd6fa457fe1296106559a3c80937a1c1069be3a3a5bd381ee6260e8d9739fce1f607',
    ),
    (
        '80808080808080808080808080808080',
        'letter advice cage absurd amount doctor acoustic avoid letter advice cage above',
        'd71de856f81a8acc65e6fc851a38d4d7ec216fd0796d0a6827a3ad6ed5511a30fa280f12eb2e47ed2ac03b5c462a0358d18d69fe4f985ec81778c1b370b652a8'
    ),
    (
        'ffffffffffffffffffffffffffffffff',
        'zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo wrong',
        'ac27495480225222079d7be181583751e86f571027b0497b5b5d11218e0a8a13332572917f0f8e5a589620c6f15b11c61dee327651a14c34e18231052e48c069'
    ),
    (
        '000000000000000000000000000000000000000000000000',
        'abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon agent',
        '035895f2f481b1b0f01fcf8c289c794660b289981a78f8106447707fdd9666ca06da5a9a565181599b79f53b844d8a71dd9f439c52a3d7b3e8a79c906ac845fa'
    ),
    (
        '7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f',
        'legal winner thank year wave sausage worth useful legal winner thank year wave sausage worth useful legal will',
        'f2b94508732bcbacbcc020faefecfc89feafa6649a5491b8c952cede496c214a0c7b3c392d168748f2d4a612bada0753b52a1c7ac53c1e93abd5c6320b9e95dd'
    ),
    (
        '808080808080808080808080808080808080808080808080',
        'letter advice cage absurd amount doctor acoustic avoid letter advice cage absurd amount doctor acoustic avoid letter always',
        '107d7c02a5aa6f38c58083ff74f04c607c2d2c0ecc55501dadd72d025b751bc27fe913ffb796f841c49b1d33b610cf0e91d3aa239027f5e99fe4ce9e5088cd65'
    ),
    (
        'ffffffffffffffffffffffffffffffffffffffffffffffff',
        'zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo when',
        '0cd6e5d827bb62eb8fc1e262254223817fd068a74b5b449cc2f667c3f1f985a76379b43348d952e2265b4cd129090758b3e3c2c49103b5051aac2eaeb890a528'
    ),
    (
        '0000000000000000000000000000000000000000000000000000000000000000',
        'abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon art',
        'bda85446c68413707090a52022edd26a1c9462295029f2e60cd7c4f2bbd3097170af7a4d73245cafa9c3cca8d561a7c3de6f5d4a10be8ed2a5e608d68f92fcc8'
    ),
    (
        '7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f',
        'legal winner thank year wave sausage worth useful legal winner thank year wave sausage worth useful legal winner thank year wave sausage worth title',
        'bc09fca1804f7e69da93c2f2028eb238c227f2e9dda30cd63699232578480a4021b146ad717fbb7e451ce9eb835f43620bf5c514db0f8add49f5d121449d3e87'
    ),
    (
        '8080808080808080808080808080808080808080808080808080808080808080',
        'letter advice cage absurd amount doctor acoustic avoid letter advice cage absurd amount doctor acoustic avoid letter advice cage absurd amount doctor acoustic bless',
        'c0c519bd0e91a2ed54357d9d1ebef6f5af218a153624cf4f2da911a0ed8f7a09e2ef61af0aca007096df430022f7a2b6fb91661a9589097069720d015e4e982f'
    ),
    (
        'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff',
        'zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo vote',
        'dd48c104698c30cfe2b6142103248622fb7bb0ff692eebb00089b32d22484e1613912f0a5b694407be899ffd31ed3992c456cdf60f5d4564b8ba3f05a69890ad'
    ),
    (
        '9e885d952ad362caeb4efe34a8e91bd2',
        'ozone drill grab fiber curtain grace pudding thank cruise elder eight picnic',
        '274ddc525802f7c828d8ef7ddbcdc5304e87ac3535913611fbbfa986d0c9e5476c91689f9c8a54fd55bd38606aa6a8595ad213d4c9c9f9aca3fb217069a41028'
    ),
    (
        '6610b25967cdcca9d59875f5cb50b0ea75433311869e930b',
        'gravity machine north sort system female filter attitude volume fold club stay feature office ecology stable narrow fog',
        '628c3827a8823298ee685db84f55caa34b5cc195a778e52d45f59bcf75aba68e4d7590e101dc414bc1bbd5737666fbbef35d1f1903953b66624f910feef245ac'
    ),
    (
        '68a79eaca2324873eacc50cb9c6eca8cc68ea5d936f98787c60c7ebc74e6ce7c',
        'hamster diagram private dutch cause delay private meat slide toddler razor book happy fancy gospel tennis maple dilemma loan word shrug inflict delay length',
        '64c87cde7e12ecf6704ab95bb1408bef047c22db4cc7491c4271d170a1b213d20b385bc1588d9c7b38f1b39d415665b8a9030c9ec653d75e65f847d8fc1fc440'
    ),
    (
        'c0ba5a8e914111210f2bd131f3d5e08d',
        'scheme spot photo card baby mountain device kick cradle pact join borrow',
        'ea725895aaae8d4c1cf682c1bfd2d358d52ed9f0f0591131b559e2724bb234fca05aa9c02c57407e04ee9dc3b454aa63fbff483a8b11de949624b9f1831a9612'
    ),
    (
        '6d9be1ee6ebd27a258115aad99b7317b9c8d28b6d76431c3',
        'horn tenant knee talent sponsor spell gate clip pulse soap slush warm silver nephew swap uncle crack brave',
        'fd579828af3da1d32544ce4db5c73d53fc8acc4ddb1e3b251a31179cdb71e853c56d2fcb11aed39898ce6c34b10b5382772db8796e52837b54468aeb312cfc3d'
    ),
    (
        '9f6a2878b2520799a44ef18bc7df394e7061a224d2c33cd015b157d746869863',
        'panda eyebrow bullet gorilla call smoke muffin taste mesh discover soft ostrich alcohol speed nation flash devote level hobby quick inner drive ghost inside',
        '72be8e052fc4919d2adf28d5306b5474b0069df35b02303de8c1729c9538dbb6fc2d731d5f832193cd9fb6aeecbc469594a70e3dd50811b5067f3b88b28c3e8d'
    ),
    (
        '23db8160a31d3e0dca3688ed941adbf3',
        'cat swing flag economy stadium alone churn speed unique patch report train',
        'deb5f45449e615feff5640f2e49f933ff51895de3b4381832b3139941c57b59205a42480c52175b6efcffaa58a2503887c1e8b363a707256bdd2b587b46541f5'
    ),
    (
        '8197a4a47f0425faeaa69deebc05ca29c0a5b5cc76ceacc0',
        'light rule cinnamon wrap drastic word pride squirrel upgrade then income fatal apart sustain crack supply proud access',
        '4cbdff1ca2db800fd61cae72a57475fdc6bab03e441fd63f96dabd1f183ef5b782925f00105f318309a7e9c3ea6967c7801e46c8a58082674c860a37b93eda02'
    ),
    (
        '066dca1a2bb7e8a1db2832148ce9933eea0f3ac9548d793112d9a95c9407efad',
        'all hour make first leader extend hole alien behind guard gospel lava path output census museum junior mass reopen famous sing advance salt reform',
        '26e975ec644423f4a4c4f4215ef09b4bd7ef924e85d1d17c4cf3f136c2863cf6df0a475045652c57eb5fb41513ca2a2d67722b77e954b4b3fc11f7590449191d'
    ),
    (
        'f30f8c1da665478f49b001d94c5fc452',
        'vessel ladder alter error federal sibling chat ability sun glass valve picture',
        '2aaa9242daafcee6aa9d7269f17d4efe271e1b9a529178d7dc139cd18747090bf9d60295d0ce74309a78852a9caadf0af48aae1c6253839624076224374bc63f'
    ),
    (
        'c10ec20dc3cd9f652c7fac2f1230f7a3c828389a14392f05',
        'scissors invite lock maple supreme raw rapid void congress muscle digital elegant little brisk hair mango congress clump',
        '7b4a10be9d98e6cba265566db7f136718e1398c71cb581e1b2f464cac1ceedf4f3e274dc270003c670ad8d02c4558b2f8e39edea2775c9e232c7cb798b069e88'
    ),
    (
        'f585c11aec520db57dd353c69554b21a89b20fb0650966fa0a9d6f74fd989d8f',
        'void come effort suffer camp survey warrior heavy shoot primary clutch crush open amazing screen patrol group space point ten exist slush involve unfold',
        '01f5bced59dec48e362f2c45b5de68b9fd6c92c6634f44d6d40aab69056506f0e35524a518034ddc1192e1dacd32c1ed3eaa3c3b131c88ed8e7e54c49a5d0998'
    )
]

class MnemonicTests(unittest.TestCase):
    def test_incorrect_length_entropy(self):
        M = Mnemonic()
        M.mnemonicFromEntropy(bytes([0x0 for i in range(32)]))
        with self.assertRaises(ValueError) as context:
            M.mnemonicFromEntropy(bytes([0x0 for i in range(13)]))
        self.assertTrue(
            'Entropy was of incorrect length.' == str(context.exception))


    def test_mnemonic_random(self):
        M = Mnemonic()
        entropy = os.urandom(32)
        phrase = M.mnemonicFromEntropy(entropy)
        assert M.entropyFromMnemonic(phrase) == entropy


    def test_lol(self):
        M = Mnemonic()
        for entropy_hex, mnemonic, seed in dataset:
            entropy = bytes.fromhex(entropy_hex)
            assert mnemonic == M.mnemonicFromEntropy(entropy)
            m = bytes(mnemonic, 'utf-8')
            assert seed == M.toSeed(m, passphrase='TREZOR')



import unittest

from bip32 import BIP32, HARDENED_INDEX


def utilTestPath(seed, path, expected_sk, expected_pk):
    bip32 = BIP32.from_seed(seed)
    sk = bip32.get_xpriv_from_path(path)
    pk = bip32.get_xpub_from_path(path)
    assert sk == expected_sk
    assert pk == expected_pk


# test cases from https://github.com/mimblewimble/grin/blob/master/keychain/src/extkey_bip32.rs#L791
def test_vector_1():
    seed = bytes.fromhex('000102030405060708090a0b0c0d0e0f')

    # m
    utilTestPath(
        seed, 'm',
        'xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi',
        'xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2gZ29ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8')

    # m/0h
    utilTestPath(
        seed, [HARDENED_INDEX],
        'xprv9uHRZZhk6KAJC1avXpDAp4MDc3sQKNxDiPvvkX8Br5ngLNv1TxvUxt4cV1rGL5hj6KCesnDYUhd7oWgT11eZG7XnxHrnYeSvkzY7d2bhkJ7',
        'xpub68Gmy5EdvgibQVfPdqkBBCHxA5htiqg55crXYuXoQRKfDBFA1WEjWgP6LHhwBZeNK1VTsfTFUHCdrfp1bgwQ9xv5ski8PX9rL2dZXvgGDnw')
    utilTestPath(
        seed, 'm/0h',
        'xprv9uHRZZhk6KAJC1avXpDAp4MDc3sQKNxDiPvvkX8Br5ngLNv1TxvUxt4cV1rGL5hj6KCesnDYUhd7oWgT11eZG7XnxHrnYeSvkzY7d2bhkJ7',
        'xpub68Gmy5EdvgibQVfPdqkBBCHxA5htiqg55crXYuXoQRKfDBFA1WEjWgP6LHhwBZeNK1VTsfTFUHCdrfp1bgwQ9xv5ski8PX9rL2dZXvgGDnw')

    # m/0h/1
    utilTestPath(
        seed, [
            HARDENED_INDEX,
            1],
        'xprv9wTYmMFdV23N2TdNG573QoEsfRrWKQgWeibmLntzniatZvR9BmLnvSxqu53Kw1UmYPxLgboyZQaXwTCg8MSY3H2EU4pWcQDnRnrVA1xe8fs',
        'xpub6ASuArnXKPbfEwhqN6e3mwBcDTgzisQN1wXN9BJcM47sSikHjJf3UFHKkNAWbWMiGj7Wf5uMash7SyYq527Hqck2AxYysAA7xmALppuCkwQ')

    # m/0h/1/2h
    utilTestPath(
        seed, [
            HARDENED_INDEX,
            1,
            HARDENED_INDEX + 2],
        'xprv9z4pot5VBttmtdRTWfWQmoH1taj2axGVzFqSb8C9xaxKymcFzXBDptWmT7FwuEzG3ryjH4ktypQSAewRiNMjANTtpgP4mLTj34bhnZX7UiM',
        'xpub6D4BDPcP2GT577Vvch3R8wDkScZWzQzMMUm3PWbmWvVJrZwQY4VUNgqFJPMM3No2dFDFGTsxxpG5uJh7n7epu4trkrX7x7DogT5Uv6fcLW5')

    # m/0h/1/2h/2
    utilTestPath(
        seed, [
            HARDENED_INDEX,
            1,
            HARDENED_INDEX + 2,
            2],
        'xprvA2JDeKCSNNZky6uBCviVfJSKyQ1mDYahRjijr5idH2WwLsEd4Hsb2Tyh8RfQMuPh7f7RtyzTtdrbdqqsunu5Mm3wDvUAKRHSC34sJ7in334',
        'xpub6FHa3pjLCk84BayeJxFW2SP4XRrFd1JYnxeLeU8EqN3vDfZmbqBqaGJAyiLjTAwm6ZLRQUMv1ZACTj37sR62cfN7fe5JnJ7dh8zL4fiyLHV')

    # m/0h/1/2h/2/1000000000
    utilTestPath(
        seed, [
            HARDENED_INDEX,
            1,
            HARDENED_INDEX + 2,
            2,
            1000000000],
        'xprvA41z7zogVVwxVSgdKUHDy1SKmdb533PjDz7J6N6mV6uS3ze1ai8FHa8kmHScGpWmj4WggLyQjgPie1rFSruoUihUZREPSL39UNdE3BBDu76',
        'xpub6H1LXWLaKsWFhvm6RVpEL9P4KfRZSW7abD2ttkWP3SSQvnyA8FSVqNTEcYFgJS2UaFcxupHiYkro49S8yGasTvXEYBVPamhGW6cFJodrTHy')


def test_vector_2():
    seed = bytes.fromhex('fffcf9f6f3f0edeae7e4e1dedbd8d5d2cfccc9c6c3c0bdbab7b4b1aeaba8a5a29f9c999693908d8a8784817e7b7875726f6c696663605d5a5754514e4b484542')

    # m
    utilTestPath(
        seed, [],
        'xprv9s21ZrQH143K31xYSDQpPDxsXRTUcvj2iNHm5NUtrGiGG5e2DtALGdso3pGz6ssrdK4PFmM8NSpSBHNqPqm55Qn3LqFtT2emdEXVYsCzC2U',
        'xpub661MyMwAqRbcFW31YEwpkMuc5THy2PSt5bDMsktWQcFF8syAmRUapSCGu8ED9W6oDMSgv6Zz8idoc4a6mr8BDzTJY47LJhkJ8UB7WEGuduB')

    # m/0
    utilTestPath(
        seed, [0],
        'xprv9vHkqa6EV4sPZHYqZznhT2NPtPCjKuDKGY38FBWLvgaDx45zo9WQRUT3dKYnjwih2yJD9mkrocEZXo1ex8G81dwSM1fwqWpWkeS3v86pgKt',
        'xpub69H7F5d8KSRgmmdJg2KhpAK8SR3DjMwAdkxj3ZuxV27CprR9LgpeyGmXUbC6wb7ERfvrnKZjXoUmmDznezpbZb7ap6r1D3tgFxHmwMkQTPH')

    # m/0/2147483647h
    utilTestPath(
        seed, [
            0,
            HARDENED_INDEX + 2147483647],
        'xprv9wSp6B7kry3Vj9m1zSnLvN3xH8RdsPP1Mh7fAaR7aRLcQMKTR2vidYEeEg2mUCTAwCd6vnxVrcjfy2kRgVsFawNzmjuHc2YmYRmagcEPdU9',
        'xpub6ASAVgeehLbnwdqV6UKMHVzgqAG8Gr6riv3Fxxpj8ksbH9ebxaEyBLZ85ySDhKiLDBrQSARLq1uNRts8RuJiHjaDMBU4Zn9h8LZNnBC5y4a')

    # m/0/2147483647h/1
    utilTestPath(
        seed, [
            0,
            HARDENED_INDEX + 2147483647,
            1],
        'xprv9zFnWC6h2cLgpmSA46vutJzBcfJ8yaJGg8cX1e5StJh45BBciYTRXSd25UEPVuesF9yog62tGAQtHjXajPPdbRCHuWS6T8XA2ECKADdw4Ef',
        'xpub6DF8uhdarytz3FWdA8TvFSvvAh8dP3283MY7p2V4SeE2wyWmG5mg5EwVvmdMVCQcoNJxGoWaU9DCWh89LojfZ537wTfunKau47EL2dhHKon')

    # m/0/2147483647h/1/2147483646h
    utilTestPath(
        seed, [
            0,
            HARDENED_INDEX + 2147483647,
            1,
            HARDENED_INDEX + 2147483646],
        'xprvA1RpRA33e1JQ7ifknakTFpgNXPmW2YvmhqLQYMmrj4xJXXWYpDPS3xz7iAxn8L39njGVyuoseXzU6rcxFLJ8HFsTjSyQbLYnMpCqE2VbFWc',
        'xpub6ERApfZwUNrhLCkDtcHTcxd75RbzS1ed54G1LkBUHQVHQKqhMkhgbmJbZRkrgZw4koxb5JaHWkY4ALHY2grBGRjaDMzQLcgJvLJuZZvRcEL')

    # m/0/2147483647h/1/2147483646h/2
    utilTestPath(
        seed, [
            0,
            HARDENED_INDEX + 2147483647,
            1,
            HARDENED_INDEX + 2147483646,
            2],
        'xprvA2nrNbFZABcdryreWet9Ea4LvTJcGsqrMzxHx98MMrotbir7yrKCEXw7nadnHM8Dq38EGfSh6dqA9QWTyefMLEcBYJUuekgW4BYPJcr9E7j',
        'xpub6FnCn6nSzZAw5Tw7cgR9bi15UV96gLZhjDstkXXxvCLsUXBGXPdSnLFbdpq8p9HmGsApME5hQTZ3emM2rnY5agb9rXpVGyy3bdW6EEgAtqt')



def test_vector_3():
    seed = bytes.fromhex('4b381541583be4423346c643850da4b320e46a87ae3d2a4e6da11eba819cd4acba45d239319ac14f863b8d5ab5a0d0c64d2e8a1e7d1457df2e5a3c51c73235be')

    # m
    utilTestPath(
        seed, [],
        'xprv9s21ZrQH143K25QhxbucbDDuQ4naNntJRi4KUfWT7xo4EKsHt2QJDu7KXp1A3u7Bi1j8ph3EGsZ9Xvz9dGuVrtHHs7pXeTzjuxBrCmmhgC6',
        'xpub661MyMwAqRbcEZVB4dScxMAdx6d4nFc9nvyvH3v4gJL378CSRZiYmhRoP7mBy6gSPSCYk6SzXPTf3ND1cZAceL7SfJ1Z3GC8vBgp2epUt13')

    # m/0h
    utilTestPath(
        seed, [HARDENED_INDEX],
        'xprv9uPDJpEQgRQfDcW7BkF7eTya6RPxXeJCqCJGHuCJ4GiRVLzkTXBAJMu2qaMWPrS7AANYqdq6vcBcBUdJCVVFceUvJFjaPdGZ2y9WACViL4L',
        'xpub68NZiKmJWnxxS6aaHmn81bvJeTESw724CRDs6HbuccFQN9Ku14VQrADWgqbhhTHBaohPX4CjNLf9fq9MYo6oDaPPLPxSb7gwQN3ih19Zm4Y')



import pytest

from mimblewimble.keychain import KeyChain


def test_key_derivation():
    master_seed = bytes.fromhex('b873212f885ccffbf4692afcb84bc2e55886de2dfa07d90f5c3c239abc31c0a6ce047e30fd8bf6a281e71389aa82d73df74c7bbfb3b06b4639a5cee775cccd3c')
    path = 'm/1/2/3/4'
    kc = KeyChain.fromSeed(master_seed)
    key1234 = kc.derivePrivateKeyAmount(path, 1234)


@pytest.mark.skip(reason="no way of currently testing this")
def test_age_encryption():
    master_seed = '2d638479b8e9ae00bee803faa7d40a8911d410bbb760d6c7da3ef52ce284cca8'
    kc = KeyChain.fromSeed(bytes.fromhex(master_seed))
    path = 'm/0/1/0'
    plaintext = bytes.fromhex('a82206e04a2e6a6c1734da52cf25695a5cac94502303dfc5eef21872801b4aff020450d285106f342cae832a0307')
    fingerprint, derived_secret, ciphertext = kc.ageEncrypt(plaintext, path)

    # without fingerprint
    decrypted = kc.ageDecrypt(ciphertext, path, derived_secret)
    assert plaintext == decrypted

    # with fingerprint
    decrypted = kc.ageDecrypt(
        ciphertext, path, derived_secret, fingerprint=fingerprint)
    assert plaintext == decrypted



import hashlib
import json

from hashlib import pbkdf2_hmac

from math import ceil

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

from mimblewimble import static


class Mnemonic:
    def __init__(self):
        # load the wordlist
        f = pkg_resources.open_binary(static, 'wordlist.json')
        self.words = json.load(f)

    def mnemonicFromEntropy(self, entropy: bytes):
        entropy_len = len(entropy)

        if entropy_len % 4 != 0:
            raise ValueError('Entropy was of incorrect length.')

        entropyBits = int(entropy_len * 8)
        checksumBits = int(entropyBits / 32)
        numWords = ceil((entropyBits + checksumBits) / 11)

        wordIndices = [0x0 for j in range(numWords)]

        loc = 0
        for i in range(entropy_len):
            byte = entropy[i]
            for j in reversed(range(1, 9)):
                bit = 1
                if (byte & (1 << (j - 1))) == 0:
                    bit = 0
                wordIndices[int(loc / 11)] |= bit << (10 - (loc % 11))
                loc += 1

        mask = ((1 << checksumBits) - 1)
        lhs = hashlib.sha256(entropy).digest()[0]
        checksum = (lhs >> (8 - checksumBits)) & mask

        for i in reversed(range(1, checksumBits+1)):
            bit = 1
            if (checksum & (1 << (i - 1))) == 0:
                bit = 0
            wordIndices[int(loc / 11)] |= bit << (10 - (loc % 11))
            loc += 1

        phrase = ''
        for i in range(numWords):
            if i > 0:
                phrase += ' '
            phrase += self.words[wordIndices[i]]

        return phrase


    def entropyFromMnemonic(self, wallet_words: str):
        words = wallet_words.split(' ')

        numWords = len(words)

        if numWords < 12 or numWords > 24 or numWords % 3 != 0:
            raise ValueError('Invalid number of words provided.')

        wordIndices = [None for i in range(numWords)]

        for i, word in enumerate(words):
            try:
                j = self.words.index(word)
            except ValueError:
                raise ValueError('Word not found: ' + word)

            wordIndices[i] = j;

        checksumBits = int(numWords / 3)
        mask = (1 << checksumBits) - 1
        expectedChecksum = wordIndices[-1] & mask;

        dataLength = int((((11 * numWords) - checksumBits) / 8) - 1)

        entropy = [0x0 for j in range(dataLength + 1)]
        entropy[-1] = wordIndices[-1] >> checksumBits

        loc = 11 - checksumBits

        for i in reversed(range(1, numWords)):
            for k in range(11):
                bit = 1
                if wordIndices[i - 1] & (1 << k) == 0:
                    bit = 0
                entropy[dataLength - int(loc / 8)] |= bit << (loc % 8)
                loc += 1

        entropy = bytes(entropy)

        lhs = hashlib.sha256(entropy).digest()[0]
        actualChecksum = (lhs >> (8 - checksumBits)) & mask

        if actualChecksum != expectedChecksum:
            raise ValueError('Invalid checksum.')

        return entropy


    def toSeed(self, entropy, passphrase=''):
        salt = bytes('mnemonic' + passphrase, 'utf-8')
        dk = pbkdf2_hmac('sha512', entropy, salt, 2048)
        return dk.hex()








import os

from mimblewimble.crypto.secret_key import SecretKey
from mimblewimble.crypto.public_key import PublicKey

from mimblewimble.crypto.aggsig import AggSig
from mimblewimble.crypto.public_keys import PublicKeys

def test_aggsig_interaction():
    agg = AggSig()
    pks = PublicKeys()

    # random message
    message = os.urandom(32)

    # sender party
    secretKeySender = SecretKey(os.urandom(32))
    publicKeySender = pks.calculatePublicKey(secretKeySender)

    secretNonceSender = agg.generateSecureNonce()
    publicNonceSender = pks.calculatePublicKey(secretNonceSender)

    # receiver party
    secretKeyReceiver = SecretKey(os.urandom(32))
    publicKeyReceiver = pks.calculatePublicKey(secretKeyReceiver)

    secretNonceReceiver = agg.generateSecureNonce()
    publicNonceReceiver = pks.calculatePublicKey(secretNonceReceiver)

    # add public keys
    sumPubKeys = pks.publicKeySum([publicKeySender, publicKeyReceiver])
    sumPubNonces = pks.publicKeySum([publicNonceSender, publicNonceReceiver])

    # generate partial signatures
    senderPartialSignature = agg.calculatePartialSignature(
        secretKeySender, secretNonceSender, sumPubKeys, sumPubNonces, message)
    receiverPartialSignature = agg.calculatePartialSignature(
        secretKeyReceiver, secretNonceReceiver, sumPubKeys, sumPubNonces, message)

    # verify partial signatures
    senderSigValid = agg.verifyPartialSignature(
        senderPartialSignature, publicKeySender,
        sumPubKeys, sumPubNonces, message)
    receiverSigValid = agg.verifyPartialSignature(
        receiverPartialSignature, publicKeyReceiver,
        sumPubKeys, sumPubNonces, message)

    assert senderSigValid
    assert receiverSigValid

    # aggregate the partial signatures
    aggregate = agg.aggregateSignatures(
        [senderPartialSignature, receiverPartialSignature], sumPubNonces)
    aggregateValid = agg.verifyAggregateSignature(aggregate, sumPubKeys, message)

    assert aggregateValid

    del agg # just to make sure secp256k1-zkp context is freed



from mimblewimble.pow.blockdb import InMemoryBlockDB
from mimblewimble.genesis import mainnet


def test_in_memory_blockdb():
    blockdb = InMemoryBlockDB()

    assert blockdb.get_block_header('nonexistent_hash') is None

    assert blockdb.get_header_by_height(0) is None
    assert blockdb.size() == 0

    blockdb.add_header(mainnet.getHeader())

    assert blockdb.get_header_by_height(0).getHash().hex() == mainnet.getHash().hex()
    assert blockdb.get_block_header(mainnet.getHash().hex()) is not None




import base58
import base64


class Codec:
    @classmethod
    def encode(self, data: bytes, encoding='utf-8') -> str:
        raise Exception('unimplemented')

    @classmethod
    def decode(self, data: str) -> bytes:
        raise Exception('unimplemented')


class HexCodec(Codec):
    @classmethod
    def encode(self, data: bytes, encoding='utf-8') -> str:
        return data.hex()

    @classmethod
    def decode(self, data: str) -> bytes:
        return bytes.fromhex(data)


class Base58Codec(Codec):
    @classmethod
    def encode(self, data: bytes, encoding='utf-8') -> str:
        return base58.b58encode(data).decode(encoding)

    @classmethod
    def decode(self, data: str) -> bytes:
        return base58.b58decode(data)


class Base64Codec(Codec):
    @classmethod
    def encode(self, data: bytes, encoding='utf-8') -> str:
        return base64.b64encode(data).decode(encoding)

    @classmethod
    def decode(self, data: str) -> bytes:
        return base64.b64decode(data)



import setuptools

setuptools.setup(
    name='mimblewimble',
    version='0.11.0',
    packages=['mimblewimble'],
    license='MIT',
    description = 'A toolset for processing Grin Mimblewimble data structures',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    author = 'Marek Narozniak',
    author_email = 'marek.yggdrasil@gmail.com',
    install_requires=[
        'base58',
        'siphash-cffi',
        'secp256k1-zkp-mw',
        'bip_utils',
        'bip32',
        'pynacl',
        'pyrage'],
    url = 'https://github.com/grinventions/mimblewimble-py',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    package_data = {
        '': ['static/__init__.py', 'static/wordlist.json'],
    }
)



import os
import pytest

from mimblewimble.models.transaction import BlindingFactor

from mimblewimble.crypto.commitment import Commitment
from mimblewimble.crypto.public_key import PublicKey
from mimblewimble.crypto.pedersen import Pedersen

# test adding blinded commitment with transparent one
def test_adding_blinded_commitments_with_transparent():
    p = Pedersen()

    blinding_factor = BlindingFactor(os.urandom(32))
    blinded_commitment = p.commit(2, blinding_factor)

    transparent_factor = BlindingFactor(bytes([0x00 for j in range(32)]))
    transparent_commitment = p.commit(3, transparent_factor)

    summed = p.commitSum([blinded_commitment, transparent_commitment], [])
    another_blinded_commitment = p.commit(5, blinding_factor)

    # they will match because of same blinding factor
    assert summed == another_blinded_commitment

# test adding 2 blinded commitment
def test_adding_two_blinded_commitment():
    p = Pedersen()

    blinding_factor_a = BlindingFactor(os.urandom(32))
    blinded_commitment_a = p.commit(2, blinding_factor_a)

    blinding_factor_b = BlindingFactor(os.urandom(32))
    blinded_commitment_b = p.commit(3, blinding_factor_b)

    summed = p.commitSum([blinded_commitment_a, blinded_commitment_b], [])

    blinding_factor_c = p.blindSum([blinding_factor_a, blinding_factor_b], [])
    blinded_commitment_c = p.commit(5, blinding_factor_c)

    # they will match because blinding factor of c is the sum
    # of blinding factors of a and b
    assert summed == blinded_commitment_c

# test adding negative blinded commitment
def test_adding_negative_blinded_commitment():
    p = Pedersen()

    blinding_factor_a = BlindingFactor(os.urandom(32))
    blinded_commitment_a = p.commit(3, blinding_factor_a)

    blinding_factor_b = BlindingFactor(os.urandom(32))
    blinded_commitment_b = p.commit(2, blinding_factor_b)

    difference = p.commitSum([blinded_commitment_a], [blinded_commitment_b])

    blinding_factor_c = p.blindSum([blinding_factor_a], [blinding_factor_b])
    blinded_commitment_c = p.commit(1, blinding_factor_c)

    # they will match because blinding factor of c is the difference
    # of blinding factors of a and b
    assert difference == blinded_commitment_c

# test public key to commitment
def test_public_key_to_commitment():
    p = Pedersen()
    public_key = PublicKey.fromHex(
        '02f434a6b929d0aa6ac757bbe387075066d51ee5308d5be91d2fb478a494d38bdf')

    commitment = p.toCommitment(public_key)
    commitment_bytes = commitment.getBytes()

    # commitment = Commitment.fromPublicKey(public_key)
    expected_commitment = Commitment.fromHex(
        '08f434a6b929d0aa6ac757bbe387075066d51ee5308d5be91d2fb478a494d38bdf')
    assert commitment == expected_commitment



from mimblewimble.slatebuilder.slate import OptionalFieldStatus


def test_optional_field_status_0():
    field_status_byte = 0
    field_status = OptionalFieldStatus.fromByte(field_status_byte)
    assert field_status.include_num_parts == 0
    assert field_status.include_amt == 0
    assert field_status.include_fee == 0
    assert field_status.include_feat == 0
    assert field_status.include_ttl == 0
    as_byte = field_status.toByte()
    assert as_byte == field_status_byte


def test_optional_field_status_2():
    field_status_byte = 2
    field_status = OptionalFieldStatus.fromByte(field_status_byte)
    assert field_status.include_num_parts == 0
    assert field_status.include_amt == 1
    assert field_status.include_fee == 0
    assert field_status.include_feat == 0
    assert field_status.include_ttl == 0
    as_byte = field_status.toByte()
    assert as_byte == field_status_byte



import unittest

from mimblewimble.models.short_id import ShortId


class ShortIdTest(unittest.TestCase):
    def test_case1(self):
        _hash = bytes.fromhex('81e47a19e6b29b0a65b9591762ce5143ed30d0261e5d24a3201752506b20f15c')
        _blockHash = (0).to_bytes(32, byteorder='big')
        _nonce = 0
        shortId = ShortId.create(_hash, _blockHash, _nonce)
        assert shortId.getId() == int.from_bytes(bytes.fromhex('4cc808b62476'), byteorder='little')


    def test_case2(self):
        _hash = bytes.fromhex('3a42e66e46dd7633b57d1f921780a1ac715e6b93c19ee52ab714178eb3a9f673')
        _blockHash = (0).to_bytes(32, byteorder='big')
        _nonce = 5
        shortId = ShortId.create(_hash, _blockHash, _nonce)
        assert shortId.getId() == int.from_bytes(bytes.fromhex('02955a094534'), byteorder='little')


    def test_case3(self):
        _hash = bytes.fromhex('3a42e66e46dd7633b57d1f921780a1ac715e6b93c19ee52ab714178eb3a9f673')
        _blockHash = bytes.fromhex('81e47a19e6b29b0a65b9591762ce5143ed30d0261e5d24a3201752506b20f15c')
        _nonce = 5
        shortId = ShortId.create(_hash, _blockHash, _nonce)
        assert shortId.getId() == int.from_bytes(bytes.fromhex('3e9cde72a687'), byteorder='little')
